import os
import sys
import argparse
import regex
import subprocess
from .style import style


def ask(question, options, formatting=None):
    """
    Takes a question for raw_input and a set of options
    that answer the question. Bugs the user until they
    choose one of the prescribed options.

    Args:
        question: The user prompt
        options (list): The responses to choose from
        formatting: A key from style_codes in clients.style

    Returns:
        response (str): The chosen `options` item.
    """

    response = ""

    get_input = input
    if sys.version_info[:2] <= (2, 7):
        get_input = raw_input
    prompt = '{} [{}] '.format(question, ', '.join(options))
    while response not in options:
        response = get_input(style(prompt, formatting))
    return response


def merge_dicts(*dicts):
    """
    Merge an arbitrary number of dicts.

    Note:
        Updates left to right, so will override existing
        attributes!
    """

    merged = {}
    for dictionary in dicts:
        merged.update(dictionary)
    return merged


def clone_pull(git_url, repo_dir):
    """
    Clones a repository from `git_url` or optionally does a 
    git pull if the repository already exists at `repo_dir`.
    """

    try:
        cmd = ['git', 'clone', git_url, repo_dir]
        subprocess.check_output(cmd, universal_newlines=True)

    except subprocess.CalledProcessError:
        response = ask('Do you want to git-pull?', ['y', 'n'], 'darkblue')
        if response == "y":
            cmd = ['git', '--git-dir=%s/.git' % (repo_dir), 'pull']
            out = subprocess.check_output(cmd, universal_newlines=True)
            print(style('Git says: {}'.format(out), 'smoke'))

    except:
        error = sys.exc_info()[0]
        print(style('Problem writing to destination: {}\n'.format(repo_dir), 'red'), error)
        raise


def execute_cmd(cmd):
    """
    Executes a command and returns the result and error output.
    """

    try:
        popen = subprocess.Popen(cmd, stderr=subprocess.PIPE, stdout=subprocess.PIPE, universal_newlines=True)
        (out, err) = popen.communicate()
    except UnicodeDecodeError:
        popen = subprocess.Popen(cmd, stderr=subprocess.PIPE, stdout=subprocess.PIPE)
        (out, err) = popen.communicate()
        try:
            out = out.decode('latin-1')
            out = out.encode('utf-8')
        except:
            error = sys.exc_info()[0]
            print(style('There was a problem executing command: {}\n'.format(cmd), 'red'), error)
            out = ''
    finally:
        return (out, err)


def parse_pre(pattern, repo_dir):
    """
    Takes a text pattern and local repo directory and returns the file name,
    line number, and text matching the given pattern in a staged revision.

    Args:
        pattern: A single text pattern to search for.
        repo_dir: The local path the repo's base directory.
    """
    pattern_re = regex.compile(pattern, regex.I)
    deleted_re = regex.compile(r'^deleted file')
    line_re = regex.compile(r'@@ \-[0-9,]+ \+([0-9]+)[, ].*')

    cmd = ['git', 'diff', '--staged', '--unified=0', '--', repo_dir]
    (out, err) = execute_cmd(cmd)

    staged_diffs = parse_diff(out, pattern_re, deleted_re, line_re)

    return staged_diffs


def parse_post(target, pattern, git_dir, revlist=None, author=None, before=None, after=None):
    """
    Takes restrictions on the elements of the revision history (message or diff)
    to search and yields a matching revision's SHA and the message or file name,
    line number, and text matching the given pattern, as well as its authorship.

    Args:
        target: The revision component being grep searched ('message' or 'diff').
        pattern: A single text pattern to search for.
        git_dir: The local path to the repo's git directory.
        revlist: The revision or revision range to be searched.
        author: (Optional) Authorship restriction on the revisions.
        before: (Option) Date restriction on the revisions.
        after: (Option) Date restriction on the revisions.

    Yields:
        sha: The SHA or a revision matching the search.
        metadata: The matching revision message or line number and text, and Authorship
            information.
    """

    pattern_re = regex.compile(r'{}'.format(pattern), regex.I)
    deleted_re = regex.compile(r'^deleted file')
    line_re = regex.compile(r'@@ \-[0-9,]+ \+([0-9]+)[, ].*')

    out = get_matching_revisions(git_dir, revlist, target, pattern, author, before, after)

    for item in out:
        sha, metadata = parse_log_metadata(item)
        if target == 'message':
            yield sha, metadata
        else:
            show_cmd = ['git', '--git-dir', git_dir, 'show', sha, '--no-color', '--unified=0']
            (out, err) = execute_cmd(show_cmd)
            file_diffs = parse_diff(out, pattern_re, deleted_re, line_re)
            if file_diffs:
                metadata['files'] = file_diffs
                yield sha, metadata


def get_matching_revisions(git_dir, revlist, target, pattern, author, before, after):

    cmd = ['git', '--git-dir', git_dir, 'log', revlist, '-i', '-E', '--oneline']

    if target == 'message':
        cmd.extend(['--format=COMMIT: %h AUTHORDATE: %aD AUTHORNAME: %an AUTHOREMAIL: %ae LOG: %s %b'])
        cmd.extend(['--grep', pattern])
    elif target == 'diff':        
        cmd.extend(['--format=COMMIT: %h AUTHORDATE: %aD AUTHORNAME: %an AUTHOREMAIL: %ae'])
        cmd.extend(['-G' + pattern])

    if author is not None:
        cmd.extend(['--author', author])
    if before is not None:
        cmd.extend(['--before', before])
    if after is not None:
        cmd.extend(['--after', after])

    (out, err) = execute_cmd(cmd)
    out = out.strip().split('COMMIT: ')[1:]
    return out


def parse_log_metadata(log):
    """
    Parses the information contained in a pretty-formatted
    --oneline commit log (i.e. the commit SHA, author's
    name and email, the date the commit was authored, and,
    optionally, the commit's message).

    Args:
        log (str): oneline pretty-formatted git log output

    Returns:
        commit (str): the revision's abbreviated SHA value.
        metadata (dict): author name, email, date (and maybe
            commit message).
    """

    metadata = {}
    sha, log = log.split(' AUTHORDATE: ', 1)
    metadata["author_date"], log = log.split(' AUTHORNAME: ', 1)
    metadata["author_name"], log = log.split(' AUTHOREMAIL: ', 1)
    try:
        metadata["author_email"], metadata["log"] = log.split(' LOG: ', 1)
    except ValueError:
        metadata["author_email"] = log
    metadata = {key: metadata[key].strip() for key in metadata.keys()}
    return sha, metadata


def parse_diff(text, pattern_re, deleted_re, line_re):
    """
    Takes a single revision and pattern. Returns the files and lines
    in the revision that match the pattern.

    Args:
        git_dir: Path to the repo's git directory.
        commit: A single revision's SHA.
        pattern_re: Compiled pattern regex.
        deleted_re: Compiled regex matching file deletions.
        line_re: Compiled regex matching line number metadata.

    Returns:
        files (list[dict]): A list of dictionaries containing the
            name of a file and any matching lines, with their text
            and line number.
    """

    files = []
    try:
        diff_list = text.split('diff --git ')[1:]
        for diff in diff_list:
            (filename, lines) = split_diff(diff, deleted_re)
            matches = [m for m in parse_diff_lines(lines, line_re, pattern_re)]
            if matches:
                files.append({"file": filename, "matches": matches})
    except TypeError:
        pass
    finally:
        return files


def split_diff(diff, deleted_re):
    """
    Divides a diff into the file name and the rest of its
    (split by newline).

    Args:
        diff: A single file's unified diff information
        deleted: Compiled regex matching deleted files

    Returns:
        If the file was not deleted in the revision:

        fname (str): The file's name
        diff_lines (list[str]): Lines modified in `diff`.
    """

    try:
        diff = diff.split('\n', 2)
        if not deleted_re.match(diff[1]):
            filename = diff[0].split(' b/', 1)[1]
            diff = '@@' + diff[2].split('@@', 1)[1]
            diff_lines = diff.split('\n')
            return filename, diff_lines
    except IndexError:
        pass


def parse_diff_lines(diff, line_re, pattern_re):
    """
    Takes the lines from a file's diff and yields them
    if they were added.

    Args:
        diff: A single file's unified diff information.
        line_re: Compiled regex matching line number information.
        pattern_re: Compiled regex matching a given pattern.

    Yields:
        line_info: A dictionary for a line where additions were made
            and which includes pattern_re. Contains the line's number
            and text.
    """

    line_num = 0
    for line in diff:
        if not line:
            pass
        elif line[0] == "@":
            line_num = int(regex.sub(line_re, r'\1', line))
        elif line[0] == "+":
            if pattern_re.search(line):
                yield {"line": line_num, "text": line[1:].strip()}
            line_num += 1




# -*- coding: utf-8 -*-
import regex


# Python character unicode

prefix = "\033["
reset = prefix + "0m"

style_codes = {
    "bold": prefix + "1m",
    "gray": prefix + "1;30m",
    "black": prefix + "30m",
    "red": prefix + "1;31m",
    "darkred": prefix + "31m",
    "green": prefix + "1;32m",
    "darkgreen": prefix + "32m",
    "yellow": prefix + "1;33m",
    "brown": prefix + "33m",
    "blue": prefix + "1;34m",
    "darkblue": prefix + "34m",
    "fuscia": prefix + "1;35m",
    "purple": prefix + "35m",
    "cyan": prefix + "1;36m",
    "darkcyan": prefix + "36m",
    "white": prefix + "1;37m",
    "smoke": prefix + "37m",
    "default": prefix + "39m",
    "yellow_bg": prefix + "43m",
    "cyan_bg": prefix + "46m",
    "blue_bg": prefix + "42m",
    "orange_bg": prefix + "41m",
    "white_bg": prefix + "47m",
    "default_bg": prefix + "49m"
}

symbol_codes = {
  'ok': u"\u2713",
  'fail': u"\u2716",
}


def style(text, _code):
    """Takes a string and styles its print output"""

    try:
        return "%s%s%s" % (style_codes[_code], text, reset)
    except:
        return "%s" % (text)


def ok(text):
    """Takes a string and prepends a green check to it"""

    return "%s %s" % (style(symbol_codes['ok'], 'green'), text)


def fail(text=''):
    """Takes a string and prepends a red x mark to it"""

    return "%s %s" % (style(symbol_codes['fail'], 'red'), text)


def highlight(text, pattern):
    """Takes a string and highlights substrings matching a pattern"""

    pattern_re = regex.compile(pattern, regex.I)
    match = pattern_re.search(text)
    if match:
        text = text.replace(match.group(0), style(match.group(0), 'red'))
    return text


def strip(text):
    return text.rstrip('\n ')


def chunk_text(text, cutoff, offset):
    cutoff = cutoff - offset
    join_str = '\n' + " "*offset
    split = text.split(' ')
    chunk = ''
    chunked = []
    while len(split):
        word = split.pop(0)
        word = word.strip(' \n')
        if not chunk and len(word) < cutoff:
            chunk = word
        elif not chunk and len(word) > cutoff:
            chunked.append(word)
        elif len(chunk) + len(word) < cutoff:
            chunk = "%s %s" % (chunk, word)
        else:
            chunked.append(chunk)
            chunk = word
    if chunk:
        chunked.append(chunk)
    output = '\n'.join(chunked).replace('\n', join_str)
    return " "*offset + output



from __future__ import print_function

import sys
import subprocess

from abc import ABCMeta, abstractmethod
from jinja2 import Environment, PackageLoader
from .style import fail, ok, highlight, style, chunk_text, strip


class AbstractClient:
    __metaclass__ = ABCMeta

    @abstractmethod
    def render(self, data):
        """
        Each instance of an AbstractClient must implement a render method, 
        which is called by Poirot
        """
        raise Exception


class ConsoleClient(AbstractClient):

    def __init__(self):
        pass

    def render(self, data, info):
        env = Environment(loader=PackageLoader('poirot', 'templates'))
        env.filters['ok'] = ok
        env.filters['fail'] = fail
        env.filters['style'] = style
        env.filters['chunk_text'] = chunk_text
        env.filters['strip'] = strip
        env.filters['highlight'] = highlight

        template = env.get_template('console.html')

        data_to_render = template.render(data=data, info=info)

        try:
            cmd = ['less', '-F', '-R', '-S', '-X', '-K']
            pager = subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=sys.stdout)
            lines = data_to_render.split('\n')
            for line in lines:
                pager.stdin.write(line.encode('utf-8') + b'\n')
            pager.stdin.close()
            pager.wait()
        except KeyboardInterrupt:
            pass
        except BrokenPipeError:
            pass

class ConsoleThinClient(AbstractClient):
    """
    Implements a minimal template for use in automated environemnts,
    e.g. continuous integration and git commit hooks
    """

    def __init__(self):
        pass

    def render(self, data, info):
        env = Environment(loader=PackageLoader('poirot', 'templates'))
        env.filters['chunk_text'] = chunk_text
        env.filters['strip'] = strip
        template = env.get_template('console_thin.html')

        data_to_render = template.render(data=data, info=info)

        print(data_to_render)



#! /usr/bin/env python
# -*- codes: utf-8 -*-

import sys
from .poirot import Poirot, Case


def poirot(args=sys.argv):
    args.pop(0)
    case = Case(args)
    poirot = Poirot(case)
    poirot.prepare()
    poirot.investigate()
    poirot.report()



try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


with open('README_PYPI.rst') as file:
    long_description = file.read()

setup(name='poirot',
	  version='0.0.21',
	  author='Emanuel Feld',
	  author_email='elefbet@gmail.com',
	  description="Investigate a Git repository's revision history to find text patterns.",
	  long_description=long_description,
	  url='https://github.com/dcgov/poirot',
	  license='https://raw.githubusercontent.com/DCgov/poirot/master/LICENSE.md',
	  packages=['poirot'],
	  install_requires=['tqdm>=3.4.0', 'Jinja2>=2.8', 'regex>=2015.11.22'],
	  test_suite='nose.collector',
	  tests_require=['nose-progressive'],
	  classifiers=[
	  	'Environment :: Console',
	  	'Intended Audience :: Developers',
	  	'Development Status :: 3 - Alpha',
        'License :: CC0 1.0 Universal (CC0 1.0) Public Domain Dedication',
        'License :: Public Domain',
	  	'Programming Language :: Python',
		'Programming Language :: Python :: 2.7',
		'Programming Language :: Python :: 3.3',
		'Programming Language :: Python :: 3.4',
		'Programming Language :: Python :: 3.5'
	  ],
	  include_package_data=True,
	  entry_points = {
	  	'console_scripts': [
	  		'poirot=poirot.commands:poirot',
	  		]
	  },
	  zip_safe=False)



from __future__ import print_function

import argparse
import os
import regex
import sys
from tqdm import tqdm

from .helpers import clone_pull, merge_dicts, parse_post, parse_pre
from .style import style
from .clients import ConsoleClient, ConsoleThinClient


class Poirot(object):

    def __init__(self, case):
        """
        Poirot needs to know who he's working for (client) and
        the details of the case. He also likes to get things in
        order by setting up a place to store findings that match
        key patterns.

        Args:
            client: ConsoleClient or ConsoleClientThin object
            case: Case object
        """

        self.case = case
        self.findings = {p: {} for p in case.patterns}

    def prepare(self):
        """
        Before conducting his investigation, Poirot must work out
        what events (revisions) to consider. He will then gather
        (clone, pull, or locate) the information to go through.
        """

        case = self.case

        if case.staged:
            print(style('Investigating staged revisions', 'blue'))

        if case.git_url:
            clone_pull(case.git_url, case.repo_dir)
        else:
            print(style('Investigating local revisions', 'blue'))
            if not os.path.exists(case.git_dir):
                raise IOError('Invalid .git directory: {}\nSpecify '
                              'the correct local directory with '
                              '--dir'.format(case.git_dir))

    def investigate(self):
        """
        Set up and initiate parsing of revisions
        in the given list of individual revisions
        or revision subsets, for the patterns
        provided.
        """

        case = self.case

        if case.staged:
            for p in case.patterns:
                self.findings[p] = parse_pre(p, case.repo_dir)

        else:
            for pattern in tqdm(case.patterns):
                for revision in case.revlist:
                    parser_args = {
                        'git_dir': case.git_dir,
                        'revlist': revision,
                        'author': case.author,
                        'before': case.before,
                        'after': case.after
                    }
                    self.parse_post('message', pattern, parser_args)
                    self.parse_post('diff', pattern, parser_args)

    def parse_post(self, item_type, pattern, parser_args):
        """
        Call commit log and diff parsers on a revision
        or revision subset and add matching commit logs
        and diffs to findings.
        """

        finding = self.findings[pattern]

        for commit, metadata in parse_post(item_type, pattern, **parser_args):
            if not hasattr(finding, commit):
                finding[commit] = {}
            finding[commit] = merge_dicts(finding[commit], metadata)

    def report(self):
        """Render findings in the console"""

        found_evidence = any(f for f in self.findings.values())
        if found_evidence:
            if self.case.verbose:
                self.client = ConsoleClient()
            else:
                self.client = ConsoleThinClient()
            self.client.render(data=self.findings, info=self.case.__dict__)
            sys.exit(1)
        else:
            print(style("Poirot didn't find anything!", 'darkblue'))
            sys.exit(0)


class Case(object):

    def __init__(self, args):
        """
        Sometimes the facts of a case come jumbled. Poirot sorts
        them out.
        """

        facts = self.parser().parse_args(args)

        self.before = facts.before
        self.after = facts.after
        self.author = facts.author
        self.staged = facts.staged
        self.verbose = facts.verbose

        self.git_url = facts.url.rstrip('/')
        self.repo_url = regex.sub(r'\.git$', '', self.git_url)
        self.repo_dir = facts.dir
        self.git_dir = facts.dir + '/.git'

        if facts.revlist == 'all':
            self.revlist = ['--all']
        else:
            self.revlist = facts.revlist.strip().split(',')

        self.patterns = set()
        if facts.term:
            self.patterns.add(facts.term)

        try:
            pfile_list = facts.patterns.strip().split(',')
            pfile_list = [pfile for pfile in pfile_list if pfile]
            for pfile in pfile_list:
                file_patterns = set([p for p in self.add_patterns(pfile)])
                self.patterns.update(file_patterns)
        except AttributeError:
            pass

        if len(self.patterns) == 0:
            print(style('No patterns given! Using default pattern set.', 'blue'))
            pattern_path = os.path.dirname(os.path.realpath(__file__))
            pattern_path = os.path.join(pattern_path, 'patterns/default.txt')
            file_patterns = set([p for p in self.add_patterns(pattern_path)])
            self.patterns.update(file_patterns)

    def add_patterns(self, file_path):
        """
        Takes a pattern file's path and yields its patterns
        """

        try:
            with open(file_path) as infile:
                for line in infile:
                    line = line.rstrip()
                    if line and not line.startswith('#'):
                        yield line
        except (IOError, OSError):
            warning = "Pattern file {} does not exist. "\
                      "Specify the correct file path with --patterns".format(file_path)
            print(style(warning, 'red'))

    def parser(self):
        query = argparse.ArgumentParser(prog='poirot', description="""Poirot:
                                        Mind Your Language""")
        query.add_argument('--url', '-u',
                           dest='url',
                           default='',
                           action="store",
                           help="""The repository's git URL,
                                e.g. 'https://github.com/dcgov/poirot.git'.""")
        query.add_argument('--dir', '-d',
                           dest='dir',
                           default=os.getcwd(),
                           help="""The path to the local directory where the
                                git repo is located or should be stored;
                                defaults to the current directory.""")
        query.add_argument('--term', '-t',
                           dest="term",
                           required=False,
                           action='store',
                           help="""A single string or regular expression
                                to search for.""")
        query.add_argument('--patterns', '-p',
                           dest="patterns",
                           action="store",
                           default=False,
                           help="""The path to the local file(s) containing strings
                                or regular expressions to match against, each
                                on a new line. Accepts a comma-separated list
                                of file paths.""")
        query.add_argument('--revlist', '-rl',
                           dest='revlist',
                           required=False,
                           default='HEAD^!',
                           help="""A comma-delimited list of revision (commit)
                                ranges to search. Defaults to HEAD^!. Specify
                                'all' to search the entire revision
                                history.""")
        query.add_argument('--before', '-b',
                           dest='before',
                           required=False,
                           help="""Search commits prior to a given date,
                                e.g., Dec-12-2015""")
        query.add_argument('--after', '-a',
                           dest='after',
                           required=False,
                           help="""Search commits after a given date,
                                e.g., Jan-01-2015""")
        query.add_argument('--author', '-au',
                           dest='author',
                           required=False,
                           help="""Restrict to commits made by an AUTHOR. An email
                                address is fine.""")
        query.add_argument('--staged', '-st',
                           dest='staged',
                           action='store_true',
                           help="""Flag to search staged modifications, instead of
                                already committed ones.""")
        query.add_argument('--verbose', '-v',
                           dest='verbose',
                           action='store_true',
                           help="""Flag to output colorful, verbose results.""")
        return query
